import numpy

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D, WavefrontDimension
from wofry.propagator.decorators import WavefrontDecorator

from LibWiser.Foundation import ComputationResults

class WiseWavefront(WavefrontDecorator):

    def __init__(self,
                 wise_computation_results=ComputationResults()):

        self.wise_computation_result = wise_computation_results

    def get_dimension(self):
        return WavefrontDimension.ONE

    def toGenericWavefront(self):
        wavelength = self.wise_computation_result.Lambda
        position = self.wise_computation_result.S
        electric_field = numpy.real(self.wise_computation_result.Field) + 1j*numpy.imag(self.wise_computation_result.Field)

        return GenericWavefront1D.initialize_wavefront_from_arrays(x_array=position, y_array=electric_field, wavelength=wavelength)

    @classmethod
    def fromGenericWavefront(cls, wavefront):
        wise_computation_result = ComputationResults()
        wise_computation_result.Lambda = wavefront.get_wavelength()
        wise_computation_result.S = wavefront.get_abscissas()
        wise_computation_result.Field = wavefront.get_complex_amplitude()

        return WiseWavefront(wise_computation_results=wise_computation_result)



from syned.beamline.optical_element import OpticalElement as SynedOpticalElement
from syned.beamline.beamline_element import BeamlineElement

from LibWiser.Foundation import OpticalElement, PositioningDirectives

class WiserOpticalElement(SynedOpticalElement):
    '''
    WiserOpticalElement acts as a generic adapter/wrapper for any optical element defined in LibWiser:
    - SourceGaussian
    - SourcePoint
    - MirrorPlane
    - MirrorElliptic
    - MirrorSpheric
    - Slits
    - Detector

    native_CoreOptics -> native_optical_element: LibWiser.Optics.<optical element>
    native_PositioningDirectives: LibWiser.Foundation.PositioningDirectives
    isSource: Flag if <optical element> is source
    native_OpticalElement: If you want to pass a ready made optical element, then leave the rest None and only pass this
    '''
    def __init__(self,
                 name="Undefined", boundary_shape=None, # Syned object
                 native_CoreOptics=None, # A LibWiser.Optics.Optics object
                 native_PositioningDirectives=PositioningDirectives(), # LibWiser Foundation PositioningDirectives
                 isSource=False,
                 native_OpticalElement=None
                 ):

        '''
        Parameters
        ---------------------
        name : str
            name of Wiser optical element
        boundary_shape: do not know
            No idea what this is...
        native_CoreOptics: LibWiser.Optics.<optical element>
            One of Wiser optical elements
        native_PositioningDirectives: LibWiser.Foundation.PositioningDirectives
            Wiser PositioningDirectives
        isSource: bool
            Source flat
        native_OpticalElement: LibWiser.Foundation.OpticalElement
            Pass the whole optical element, leave native_CoreOptics=None and native_PositioningDirectives=None
        '''

        # Init SynedOpticalElement
        # Fills: name, boundary_shape, as required by Syned
        super(WiserOpticalElement, self).__init__(name=name, boundary_shape=boundary_shape)

        if native_CoreOptics is not None:
            # Build the native Wiser OpticalElement
            # fills native_optical_element
            self.native_optical_element = OpticalElement(CoreOpticsElement=native_CoreOptics,
                                                         PositioningDirectives=native_PositioningDirectives,
                                                         Name=name,
                                                         # syned name and Wiser name will never be automatically synced...
                                                         IsSource=isSource)  # In case of source, to be discussed
        elif native_OpticalElement is not None:
            # In case the whole optical element is passed as an argument
            self.native_optical_element = native_OpticalElement
        else:
            print('Native OpticalElement not specified!')


class WiserBeamlineElement(BeamlineElement):
    '''
    Glue for beamline element. Nothing particular here.
    '''
    def __init__(self, optical_element=WiserOpticalElement()):
        super(WiserBeamlineElement, self).__init__(optical_element=optical_element, coordinates=None)

    def get_coordinates(self):
        raise NotImplementedError("this method cannot be used in WISE 2")



# THIS FILE IS GENERATED FROM wofrywise SETUP.PY
short_version = '0.0.8'
version = '0.0.8'
full_version = '0.0.8'
git_revision = 'f0ce78aee7305610f8c98e563a17362ea2eda595'
release = True

if not release:
    version = full_version
    short_version += ".dev"



import numpy

import scipy.constants as codata
angstroms_to_eV = codata.h*codata.c/codata.e*1e10

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from wofry.propagator.propagator import Propagator1D, PropagationParameters, PropagationElements, PropagationManager, PropagationMode

from WofryWiser.propagator.wavefront1D.wise_wavefront import WiseWavefront
from WofryWiser.beamline.beamline_elements import WiserBeamlineElement

from LibWiser import Foundation, Optics

WISE_APPLICATION = "All"

class WisePropagationElements(PropagationElements):

    __wise_propagation_elements = None

    def __init__(self):
        super(WisePropagationElements, self).__init__()

        self.__wise_propagation_elements = Foundation.BeamlineElements()

    def add_beamline_element(self, beamline_element=WiserBeamlineElement()):
        super(WisePropagationElements, self).add_beamline_element(beamline_element)

        self.__wise_propagation_elements.Append(beamline_element.get_optical_element().native_optical_element)

    def insert_beamline_element(self, index, new_element=WiserBeamlineElement(), mode=PropagationElements.INSERT_BEFORE):

        existing_element_name = self.get_wise_propagation_element(index).Name  # Has to be called before calling super()

        super(WisePropagationElements, self).insert_beamline_element(index, new_element, mode)

        self.__wise_propagation_elements.Insert(NewItem=new_element.get_optical_element().native_optical_element,
                                                ExistingName=existing_element_name,
                                                Mode=1)

    def add_beamline_elements(self, beamline_elements=[]):
        for beamline_element in beamline_elements:
            self.add_beamline_element(beamline_element)

    def get_wise_propagation_element(self, index):
        return self.get_propagation_element(index).get_optical_element().native_optical_element

    def get_wise_propagation_elements(self):
        return self.__wise_propagation_elements

    def refresh_wise_positions(self):
        self.get_wise_propagation_elements().RefreshPositions()

class WisePropagator(Propagator1D):

    HANDLER_NAME = "WISE2_PROPAGATOR"

    def get_handler_name(self):
        return self.HANDLER_NAME

    def do_propagation(self, parameters=PropagationParameters()):
        wavefront = parameters.get_wavefront()

        if not wavefront is None:
            is_generic_wavefront = isinstance(wavefront, GenericWavefront1D)
        else:
            is_generic_wavefront = False

        if not is_generic_wavefront and not wavefront is None:
            if not isinstance(wavefront, WiseWavefront): raise ValueError("Wavefront cannot be managed by this propagator")

        if is_generic_wavefront:
            wavefront = WiseWavefront.fromGenericWavefront(wavefront)

        wise_propagation_elements = parameters.get_PropagationElements()

        beamline = wise_propagation_elements.get_wise_propagation_elements()
        beamline.ComputationSettings.NPools = int(parameters.get_additional_parameter("NPools"))

        # Element -1 is the last element of the array
        optical_element_end = wise_propagation_elements.get_propagation_element(-1).get_optical_element()

        oeEnd = optical_element_end.native_optical_element
        if parameters.get_additional_parameter("single_propagation"):
            oeStart = oeEnd.GetParent(SameOrientation=True, OnlyReference=False)
        else:
            oeStart = wise_propagation_elements.get_wise_propagation_element(0)

        # oeEnd = optical_element_end.native_optical_element
        # oeStart is the first element [0], or the previous element [-2]
        # print('+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        # print(wise_propagation_elements.get_wise_propagation_element(-2 if parameters.get_additional_parameter("single_propagation") else 0).Name, type(wise_propagation_elements.get_wise_propagation_element(-2 if parameters.get_additional_parameter("single_propagation") else 0)))
        # print(oeStart.Name, oeEnd.Name)
        # print(oeStart.CoreOptics.Orientation, oeEnd.CoreOptics.Orientation)
        # print('+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
        # oeStart =wise_propagation_elements.get_wise_propagation_element(-2 if parameters.get_additional_parameter("single_propagation") else 0)# oeEnd.GetParent(SameOrientation=True)

        if PropagationManager.Instance().get_propagation_mode(WISE_APPLICATION) == PropagationMode.STEP_BY_STEP or parameters.get_additional_parameter("is_full_propagator"):

            beamline.RefreshPositions()
            beamline.ComputeFields(oeStart=oeStart, oeEnd=oeEnd, Verbose=False)

            result = WiseWavefront(wise_computation_results=oeEnd.ComputationData)
        elif PropagationManager.Instance().get_propagation_mode(WISE_APPLICATION) == PropagationMode.WHOLE_BEAMLINE:
            result = wavefront
        else:
            result = None

        if is_generic_wavefront:
            return None if result is None else result.toGenericWavefront()
        else:
            return result





# -*- coding: utf-8 -*-
"""
Test script - build a dpi horizontal beamline.
"""
from wofry.propagator.propagator import PropagationParameters, PropagationManager

from WofryWiser.beamline.beamline_elements import WiserBeamlineElement, WiserOpticalElement
from WofryWiser.propagator.wavefront1D.wise_wavefront import WiseWavefront
from WofryWiser.propagator.propagator1D.wise_propagator import WisePropagator, WisePropagationElements

from LibWiser.WiserImport import *
import LibWiser.FermiSource as FermiSource

Distances = FermiSource.DistancesF2
PathMetrologyFermi = Paths.MetrologyFermi

UseGrating = False
Set = 0
Lambda = 2e-9
Waist0 = 180e-6

def plot(oe):
    ToolLib.CommonPlots.IntensityAtOpticalElement(oe)

if Set ==0:
	Angle_pm2a_deg = 2.5
	Angle_presto_deg = 2.5
	Angle_kb_deg = 2.0

elif Set  == 1:
	Angle_pm2a_deg = 2.0
	Angle_presto_deg = 2.0
	Angle_kb_deg = 2.0

elif Set  == 2:
	Angle_pm2a_deg = 2.0
	Angle_presto_deg = 2.0
	Angle_kb_deg = 2.

elif Set == 3:
	Angle_pm2a_deg = 2.5
	Angle_presto_deg = 2.5
	Angle_kb_deg = 2.0

a = np.array([0])
tl.Debug.On = True

DetectorSize = 100e-6


print(__name__)
if __name__ == '__main__':

    PropagationManager.Instance().add_propagator(WisePropagator())

    # SOURCE (H, V)
    # -----------------------------------------------------------------------------
    DeltaSource = 0
    ww_s = WiserOpticalElement(name='FEL2',
                               boundary_shape=None,
                               isSource=True,
                               native_CoreOptics=Optics.SourceGaussian(
                                   Lambda=Lambda,
                                   Waist0=Waist0,
                                   Orientation=Optics.OPTICS_ORIENTATION.HORIZONTAL),
                               native_PositioningDirectives=Foundation.PositioningDirectives(
                                   ReferTo='absolute',
                                   XYCentre=[0,0],
                                   Angle=0.)
                               )

    s = ww_s.native_optical_element # Wiser optical element


    # PM2A (H)
    # -----------------------------------------------------------------------------

    ww_pm2a = WiserOpticalElement(name='PM2a',
                                  boundary_shape=None,
                                  native_CoreOptics=Optics.MirrorPlane(
                                      L=0.4,
                                      AngleGrazing=np.deg2rad(Angle_pm2a_deg),
                                      Orientation=Optics.OPTICS_ORIENTATION.HORIZONTAL),
                                  native_PositioningDirectives=Foundation.PositioningDirectives(
                                      ReferTo='source',
                                      PlaceWhat='centre',
                                      PlaceWhere='centre',
                                      Distance=41.4427)
                                  )

    pm2a = ww_pm2a.native_optical_element

    # Presto (H)
    # -----------------------------------------------------------------------------

    ww_presto = WiserOpticalElement(name='presto',
                                    boundary_shape=None,
                                    native_CoreOptics=Optics.MirrorPlane(
                                        L=0.4,
                                        AngleGrazing=np.deg2rad(Angle_presto_deg),
                                        Orientation=Optics.OPTICS_ORIENTATION.HORIZONTAL),
                                    native_PositioningDirectives=Foundation.PositioningDirectives(
                                        ReferTo='source',
                                        PlaceWhat='centre',
                                        PlaceWhere='centre',
                                        Distance=49.8466)
                                    )

    presto = ww_presto.native_optical_element

    # dpi kb (H)
    # -----------------------------------------------------------------------------

    ww_dpi_kbh = WiserOpticalElement(name='kb',
                                     boundary_shape=None,
                                     native_CoreOptics=Optics.MirrorElliptic(
                                         L=0.4,
                                         f1=91.55,
                                         f2=1.2,
                                         Alpha=np.deg2rad(Angle_kb_deg),
                                         Orientation=Optics.OPTICS_ORIENTATION.HORIZONTAL),
                                     native_PositioningDirectives=Foundation.PositioningDirectives(
                                         ReferTo='source',
                                         PlaceWhat='centre',
                                         PlaceWhere='centre',
                                         Distance=91.55)
                                     )

    dpi_kbh = ww_dpi_kbh.native_optical_element

    # Detector (H)
    # -----------------------------------------------------------------------------

    ww_d_h = WiserOpticalElement(name='d_h',
                                 boundary_shape=None,
                                 native_CoreOptics=Optics.Detector(
                                     L=DetectorSize,
                                     AngleGrazing=np.deg2rad(90),
                                     Orientation=Optics.OPTICS_ORIENTATION.HORIZONTAL),
                                 native_PositioningDirectives=Foundation.PositioningDirectives(
                                     ReferTo='upstream',
                                     PlaceWhat='centre',
                                     PlaceWhere='centre',
                                     Distance=1.2)
                                 )

    d_h = ww_d_h.native_optical_element

    # Create beamline
    # -----------------------------------------------------------------------------

    th = WisePropagationElements()
    th.get_wise_propagation_elements().ComputationSettings.OrientationToCompute = [Optics.OPTICS_ORIENTATION.HORIZONTAL]
    th.add_beamline_element(WiserBeamlineElement(optical_element=ww_s))
    th.add_beamline_element(WiserBeamlineElement(optical_element=ww_pm2a))
    th.add_beamline_element(WiserBeamlineElement(optical_element=ww_presto))
    th.add_beamline_element(WiserBeamlineElement(optical_element=ww_dpi_kbh))
    th.add_beamline_element(WiserBeamlineElement(optical_element=ww_d_h))

    th.refresh_wise_positions()
    print(th.get_wise_propagation_elements())

    # Computation settings
    # -----------------------------------------------------------------------------
    NSampling = int(5e3)

    pm2a.ComputationSettings.UseCustomSampling = True
    pm2a.ComputationSettings.NSamples = NSampling

    presto.ComputationSettings.UseCustomSampling = True
    presto.ComputationSettings.NSamples = NSampling

    dpi_kbh.ComputationSettings.UseCustomSampling = True
    dpi_kbh.ComputationSettings.NSamples = NSampling

    d_h.ComputationSettings.UseCustomSampling = True
    d_h.ComputationSettings.NSamples = NSampling

    # Surface settings
    # -----------------------------------------------------------------------------
    # pm2a
    x, y, FileInfo = tl.Metrology.ReadLtp2File(PathMetrologyFermi / "PM2A" / "PM2A_I06.SLP")  # read slopes
    h = tl.Metrology.SlopeIntegrate(y, dx=FileInfo.XStep)  # integrate slopes
    pm2a.CoreOptics.FigureErrorLoad(h,
                                    Step=FileInfo.XStep,
                                    AmplitudeScaling=1,  # => m  UNDERSTAND IF -1 or +1
                                    Append=False)
    pm2a.CoreOptics.ComputationSettings.UseFigureError = True

    # presto
    h = tl.FileIO.ReadYFile(PathMetrologyFermi / "HE_WISE_FigureError.txt", SkipLines=1)
    h = h*1e-3 # mm => m
    if UseGrating:
        hGrating, GratingStep = tl.Metrology.RectangularGrating(
            L0=250e-3,
            L1=60e-3,
            N=3e5,
            LinesPerMillimiter=3750,
            GrooveHeight=9.5e-9,
            HighDuty=0.65,
            ReturnStep=True)
        NGrating = len(hGrating)
        hNew = rm.FastResample1d(h, NGrating) # FigureError and Grating profiles have equal sampling

        hTot = hNew + hGrating
        XStep = GratingStep

    else:
        hTot = h
        XStep = 1e-3

    presto.CoreOptics.FigureErrorLoad(hTot, Step=XStep, AmplitudeScaling=1, Append=False)

    # kbh
    dpi_kbh.CoreOptics.ComputationSettings.UseFigureError = True      # use figure error?
    dpi_kbh.CoreOptics.FigureErrorLoad(
        File=PathMetrologyFermi / "dpi_kbh.txt",
        Step=2e-3, # passo del file
        AmplitudeScaling=-1 * 1e-3,  # fattore di scala
        Append=False
    )

    # Ignore settings
    # -----------------------------------------------------------------------------
    # Ignore List
    pm2a.ComputationSettings.Ignore = False
    presto.ComputationSettings.Ignore = False
    dpi_kbh.ComputationSettings.Ignore = False
    d_h.ComputationSettings.Ignore = False

    UseFigureError = True
    pm2a.CoreOptics.ComputationSettings.UseFigureError = UseFigureError
    presto.CoreOptics.ComputationSettings.UseFigureError = UseFigureError
    dpi_kbh.CoreOptics.ComputationSettings.UseFigureError = UseFigureError

    # Initialize computation parameters
    # -----------------------------------------------------------------------------

    wavefront = WiseWavefront(wise_computation_results=None)
    parameters = PropagationParameters(wavefront=WiseWavefront(wise_computation_results=None), propagation_elements=th)
    parameters.set_additional_parameters("single_propagation", False)
    parameters.set_additional_parameters("NPools", 1)

    wavefront = PropagationManager.Instance().do_propagation(propagation_parameters=parameters,
                                                             handler_name=WisePropagator.HANDLER_NAME)
    # Plot results
    # -----------------------------------------------------------------------------

    plot(d_h)
    plt.show()



# -*- coding: utf-8 -*-
"""
Test script to try whether adding and removing an optical element still results in correct calculations.
"""
import PyQt5

import importlib

import LibWiser.Rayman as rm
import LibWiser.Foundation as Foundation
import LibWiser.Optics as Optics
import LibWiser.ToolLib as tl
import LibWiser.FermiSource as Fermi

importlib.reload(Foundation)
importlib.reload(Optics)
importlib.reload(tl)
importlib.reload(rm)
importlib.reload(Fermi)

from matplotlib import pyplot as plt

from numpy import *
import numpy as np

from LibWiser import ToolLib as tl
from wofry.propagator.propagator import PropagationParameters, PropagationManager

from WofryWiser.beamline.beamline_elements import WiserBeamlineElement
from WofryWiser.beamline.beamline_elements import WiserOpticalElement

from WofryWiser.propagator.wavefront1D.wise_wavefront import WiseWavefront
from WofryWiser.propagator.propagator1D.wise_propagator import WisePropagator, WisePropagationElements

def plot(oe):
    tl.CommonPlots.IntensityAtOpticalElement(oe)

# PARAMETERS

Lambda = 2e-9
Waist0 = 180e-6

f1_h = Fermi.Dpi.Kbh.f1
f2_h = Fermi.Dpi.Kbh.f2
M_h = f1_h / f2_h
f1_v = Fermi.Dpi.Kbv.f1
f2_v = Fermi.Dpi.Kbv.f2
M_v = f1_v / f2_v

L = 0.4
UseFigureError = True

NPools = 1
DetectorSize = 150e-6
DetectorRestPosition = 0

DoCaustics = True
NDeltaFocus = 10
RangeOfDeltaFocus = 10e-3  # estensione del delta focus
DeltaFocusOffset = 0e-3  # dove centro il delta focus (se cambio la sorgente, dovrebbe cambiare)

DeltaSourceList = np.zeros(1)  # espressa in unità di ondulatori

AngleGrazing = deg2rad(2.)

single_time = []

print(__name__)
if __name__ == '__main__':

    PropagationManager.Instance().add_propagator(WisePropagator())

    tl.Debug.On = True
    N = 3000
    UseCustomSampling = True
    # SOURCE
    #==========================================================================
    Lambda = 2e-9
    Waist0 = 180e-6#Fermi.Waist0E(Lambda)

    ww_s = WiserOpticalElement(name='FEL2 source',
                               boundary_shape=None,
                               isSource=True,
                               native_CoreOptics=Optics.SourceGaussian(Lambda=Lambda, Waist0=Waist0),
                               native_PositioningDirectives=Foundation.PositioningDirectives(
                                   ReferTo='absolute',
                                   XYCentre=[0, 0],
                                   Angle=pi/4.-AngleGrazing)
                               )

    s = ww_s.native_optical_element # Wiser optical element
    s.ComputationSettings.UseSmallDisplacements = True
    s.Orientation = Optics.OPTICS_ORIENTATION.ANY
    s.CoreOptics.SmallDisplacements.Long = DeltaSourceList[0]
    s.CoreOptics.SmallDisplacements.Rotation = 0.

    # PM1 (h)
    #==========================================================================
    ww_pm1_h = WiserOpticalElement(name='pm1h',
                                   boundary_shape=None,
                                   native_CoreOptics=Optics.MirrorPlane(L=L, AngleGrazing=AngleGrazing),
                                   native_PositioningDirectives=Foundation.PositioningDirectives(
                                       ReferTo='upstream',
                                       PlaceWhat='centre',
                                       PlaceWhere='centre',
                                       Distance=20.)
                                   )

    pm1_h = ww_pm1_h.native_optical_element # Wiser optical element
    pm1_h.ComputationSettings.UseFigureError = False
    pm1_h.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.HORIZONTAL

    pm1_h.ComputationSettings.Ignore = False          # Lo user decide di non simulare lo specchio ()
    pm1_h.ComputationSettings.UseCustomSampling = UseCustomSampling # l'utente decide di impostare a mano il campionamento
    pm1_h.ComputationSettings.NSamples = N

    # PM2 (h)
    # ==========================================================================
    ww_pm2_h = WiserOpticalElement(name='pm2h',
                                   boundary_shape=None,
                                   native_CoreOptics=Optics.MirrorPlane(L=L, AngleGrazing=AngleGrazing),
                                   native_PositioningDirectives=Foundation.PositioningDirectives(
                                       ReferTo='upstream',
                                       PlaceWhat='centre',
                                       PlaceWhere='centre',
                                       Distance=20.)
                                   )

    pm2_h = ww_pm2_h.native_optical_element # Wiser optical element
    pm2_h.ComputationSettings.UseFigureError = False
    pm2_h.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.HORIZONTAL

    pm2_h.ComputationSettings.Ignore = False  # Lo user decide di non simulare lo specchio ()
    pm2_h.ComputationSettings.UseCustomSampling = UseCustomSampling  # l'utente decide di impostare a mano il campionamento
    pm2_h.ComputationSettings.NSamples = N

    # KB(v)
    # ==========================================================================
    ww_kb_v = WiserOpticalElement(name='kb_v',
                                  boundary_shape=None,
                                  native_CoreOptics=Optics.MirrorElliptic(f1=f1_v, f2=f2_v, L=L, Alpha=AngleGrazing),
                                  native_PositioningDirectives=Foundation.PositioningDirectives(
                                      ReferTo='source',
                                      PlaceWhat='upstream focus',
                                      PlaceWhere='centre')
                                  )

    kb_v = ww_kb_v.native_optical_element
    kb_v.CoreOptics.ComputationSettings.UseFigureError = False
    kb_v.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.VERTICAL
    kb_v.CoreOptics.ComputationSettings.UseRoughness = False
    kb_v.CoreOptics.ComputationSettings.UseSmallDisplacements = False  # serve per traslare/ruotare l'EO

    kb_v.ComputationSettings.UseCustomSampling = UseCustomSampling  # l'utente decide di impostare a mano il campionamento
    kb_v.ComputationSettings.NSamples = N
    # kb_v.CoreOptics.FigureErrorLoad(File='DATA/LTP/kbh.txt', Step=2e-3, AmplitudeScaling=-1e-3)

    # KB(h)
    # ==========================================================================
    ww_kb_h = WiserOpticalElement(name='kb_h',
                                  boundary_shape=None,
                                  native_CoreOptics=Optics.MirrorElliptic(f1=f1_h, f2=f2_h, L=L, Alpha=AngleGrazing),
                                  native_PositioningDirectives=Foundation.PositioningDirectives(
                                      ReferTo='source',
                                      PlaceWhat='upstream focus',
                                      PlaceWhere='centre')
                                  )

    kb_h = ww_kb_h.native_optical_element
    kb_h.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.HORIZONTAL
    kb_h.CoreOptics.ComputationSettings.UseRoughness = False
    kb_h.CoreOptics.ComputationSettings.UseSmallDisplacements = False  # serve per traslare/ruotare l'EO
    kb_h.ComputationSettings.UseCustomSampling = UseCustomSampling  # l'utente decide di impostare a mano il campionamento
    kb_h.ComputationSettings.NSamples = N

    kb_h.CoreOptics.ComputationSettings.UseFigureError = True
    kb_h.CoreOptics.FigureErrorLoad(File='kbh.txt', Step=2e-3, AmplitudeScaling=-1e-3)

    ww_sl = WiserOpticalElement(name='sl',
                                boundary_shape=None,
                                native_CoreOptics=Optics.Slits(L=1e-3),
                                native_PositioningDirectives=Foundation.PositioningDirectives(
                                    ReferTo='source',
                                    PlaceWhat='centre',
                                    PlaceWhere='centre',
                                    Distance=f2_v + f1_v - 0.5),
                                )

    ww_sl.native_optical_element.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.VERTICAL
    ww_sl.native_optical_element.CoreOptics.UseAsReference = False

    # detector (v)
    # ==========================================================================
    ww_d_v = WiserOpticalElement(name='detector_v',
                                 boundary_shape=None,
                                 native_CoreOptics=Optics.Detector(L=DetectorSize, AngleGrazing=deg2rad(90)),
                                 native_PositioningDirectives=Foundation.PositioningDirectives(
                                     ReferTo='upstream',
                                     PlaceWhat='centre',
                                     PlaceWhere='downstream focus')
                                 )

    d_v = ww_d_v.native_optical_element
    d_v.ComputationSettings.UseCustomSampling = UseCustomSampling
    d_v.ComputationSettings.NSamples = N                          # come sopra. In teoria il campionamento può essere specificato elemento per elmeento
    d_v.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.VERTICAL

    # detector (h)
    # ==========================================================================
    ww_d_h = WiserOpticalElement(name='detector_h',
                                 boundary_shape=None,
                                 native_CoreOptics=Optics.Detector(L=DetectorSize, AngleGrazing=deg2rad(90)),
                                 native_PositioningDirectives=Foundation.PositioningDirectives(
                                     ReferTo='upstream',
                                     PlaceWhat='centre',
                                     PlaceWhere='downstream focus')
                                 )

    d_h = ww_d_h.native_optical_element
    d_h.ComputationSettings.UseCustomSampling = UseCustomSampling
    d_h.ComputationSettings.NSamples = N                          # come sopra. In teoria il campionamento può essere specificato elemento per elmeento
    d_h.CoreOptics.Orientation = Optics.OPTICS_ORIENTATION.HORIZONTAL

    beamline = WisePropagationElements()
    beamline.get_wise_propagation_elements().ComputationSettings.OrientationToCompute = [Optics.OPTICS_ORIENTATION.HORIZONTAL, Optics.OPTICS_ORIENTATION.VERTICAL]

    wavefront = WiseWavefront(wise_computation_results=None)

    # Build the beamline
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_s))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_pm1_h))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_kb_v))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_kb_h))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_sl))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_d_v))
    beamline.add_beamline_element(WiserBeamlineElement(optical_element=ww_d_h))

    parameters = PropagationParameters(wavefront=WiseWavefront(wise_computation_results=None), propagation_elements=beamline)
    parameters.set_additional_parameters("single_propagation", False)
    parameters.set_additional_parameters("NPools", 1)
    #
    wavefront = PropagationManager.Instance().do_propagation(propagation_parameters=parameters, handler_name=WisePropagator.HANDLER_NAME)
    #
    beamline_before = beamline
    # # Plot the result

    plot(d_v)

    plot(d_h)
    #
    # # Insert an element
    beamline.insert_beamline_element(index=1, new_element=WiserBeamlineElement(optical_element=ww_pm2_h))

    parameters = PropagationParameters(wavefront=WiseWavefront(wise_computation_results=None),
                                       propagation_elements=beamline)
    parameters.set_additional_parameters("single_propagation", False)
    parameters.set_additional_parameters("NPools", 1)
    #
    wavefront = PropagationManager.Instance().do_propagation(propagation_parameters=parameters,
                                                             handler_name=WisePropagator.HANDLER_NAME)

    print('Original beamline')
    # print(beamline.get_wise_propagation_elements())
    print(beamline_before.get_wise_propagation_elements())

    plot(d_v)
    plot(d_h)

    print('Beamline with inserted element')
    print(beamline.get_wise_propagation_elements())



    plt.show()
